(*
  Fibonacci using TAL_t

  Nadeem Abdul Hamid, February 2006
*)

Require Import List.
Require Import tactics.
Require Import talt.

Definition halt_addr := 0.
Definition fib_entry_addr := 1.
Definition fib_body_addr := 2.

Definition halt_type : ctype :=
  cbind (
  cbindC 1 (
  cbindC 2 (
  cbindC 3 (
  cbindC 4 (
  cbindC 5 (
  cbindC 6 (
    cbaseC _ (
    rftconsC _ r0 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tliftC _ 
                           (tliftC _ (tliftC _ tintC))))))) (
    rftconsC _ r1 (tvarC 6) (
    rftconsC _ r2 (tliftC _ (tvarC 5)) (
    rftconsC _ r3 (tliftC _ (tliftC _ (tvarC 4))) (
    rftconsC _ r4 (tliftC _ (tliftC _ (tliftC _ (tvarC 3)))) (
    rftconsC _ r5 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tvarC 2))))) (
    rftconsC _ r6 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tvarC 1)))))) (
    rftconsC _ r7 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tvarC 0))))))) (
    (rftnilC _))))))))))))))))).

Definition halt_ciseq : ciseq :=
  cibind (fun a1 =>
  cibind (fun a2 =>
  cibind (fun a3 =>
  cibind (fun a4 =>
  cibind (fun a5 =>
  cibind (fun a6 =>
  cibind (fun a7 =>
    cibase (
      ijd halt_addr (a1::a2::a3::a4::a5::a6::a7::nil)
    )))))))).




Definition fib_entry_type : ctype :=
  cbind (
  cbindC 1 (
  cbindC 2 (
  cbindC 3 (
    cbaseC 4 (
      rftconsC 4 r0 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tintC))))) (
      rftconsC 4 r1 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tintC))))) (
      rftconsC 4 r2 (tliftC _ (tliftC _ (tliftC _ (tliftC _ (tintC))))) (
      rftconsC 4 r3 (tvarC 3) (
      rftconsC 4 r4 (tliftC _ (tvarC 2)) (
      rftconsC 4 r5 (tliftC _ (tliftC _ (tvarC 1))) (
      rftconsC 4 r6 (tliftC _ (tliftC _ (tliftC _ (tvarC 0)))) (
      rftconsC 4 r7 (tliftC _ (tliftC _ (tliftC _ (tliftC _
	(lift (tcode halt_type))))))
      (rftnilC _))))))))))))).

Definition fib_entry_ciseq : ciseq :=
  cibind (fun a3 => 
  cibind (fun a4 => 
  cibind (fun a5 => 
  cibind (fun a6 => 
    cibase (
      icons (ibgti r0 0 fib_body_addr (a3::a4::a5::a6::nil)) (
      icons (imov r0 r1) (
      icons (itapp r7 tint) (
      icons (itapp r7 tint) (
      icons (itapp r7 a3) (
      icons (itapp r7 a4) (
      icons (itapp r7 a5) (
      icons (itapp r7 a6) (
      icons (itapp r7 ttop) (
	ijmp r7
      )))))))))))))).


Definition fib_body_type : ctype := fib_entry_type.


Definition fib_body_ciseq : ciseq :=
  cibind (fun a3 => 
  cibind (fun a4 => 
  cibind (fun a5 => 
  cibind (fun a6 => 
    cibase (
      icons (imov r3 r1) (
      icons (imov r1 r2) (
      icons (iadd r2 r2 r3) (
      icons (isubi r0 r0 1) (
	ijd fib_entry_addr (tint::a4::a5::a6::nil)
      ))))))))).


Definition CH : codeheap :=
  fun a =>
    match a with
      | 0 => Some (halt_type, halt_ciseq)
      | 1 => Some (fib_entry_type, fib_entry_ciseq)
      | 2 => Some (fib_body_type, fib_body_ciseq)
      | _ => None
    end.


Lemma wf_halt : (wf_codeval CH (halt_type, halt_ciseq)).
Proof.
  constructor; intros.
  repeat mcinvClear app_ctype.
  unfold halt_ciseq.
  rewrite (unlift_lift_eq t).
  rewrite (unlift_lift_eq t0).
  rewrite (unlift_lift_eq t1).
  rewrite (unlift_lift_eq t2).
  rewrite (unlift_lift_eq t3).
  rewrite (unlift_lift_eq t4).
  rewrite (unlift_lift_eq t5).

  exists (ijd halt_addr (t::t0::t1::t2::t3::t4::t5::nil)).
  constructor.
  repeat constructor.
  eapply wf_ijd; eauto.
  compute; eauto.
  unfold ctypeOf, fst, halt_type.
  repeat (constructor; auto; compute).
  generalize (unlift_lift_eq t); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t0); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t1); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t2); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t3); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t4); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t5); compute; intros Ht; rewrite Ht; clear Ht.
  constructor.
Qed.


Lemma wf_fib_entry : (wf_codeval CH (fib_entry_type, fib_entry_ciseq)).
Proof.
  constructor; intros.
  repeat mcinvClear app_ctype.
  rewrite (unlift_lift_eq t).
  rewrite (unlift_lift_eq t0).
  rewrite (unlift_lift_eq t1).
  rewrite (unlift_lift_eq t2).
  unfold fib_entry_ciseq.

  exists 
    (icons (ibgti r0 0 fib_body_addr (t :: t0 :: t1 :: t2 :: nil))
      (icons (imov r0 r1)
	(icons (itapp r7 tint)
          (icons (itapp r7 tint) 
	    (icons (itapp r7 t)
	    (icons (itapp r7 t0)
	    (icons (itapp r7 t1)
	    (icons (itapp r7 t2)
	    (icons (itapp r7 ttop)
	      (ijmp r7)))))))))).
  split; auto.
  repeat constructor.

  (* bgti *)
  eapply wf_ibgti; eauto.
  compute; eauto.
  compute;
  repeat (constructor; auto; compute).
  generalize (unlift_lift_eq t); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t0); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t1); compute; intros Ht; rewrite Ht; clear Ht.
  generalize (unlift_lift_eq t2); compute; intros Ht; rewrite Ht; clear Ht.
  constructor; auto.

  (* mov *)
  constructor; auto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.
  eapply wf_itapp; eauto. compute; eauto.

  (* jmp *)
  eapply wf_ijmp; eauto.
  compute; eauto.

  intro r; elim r; compute; auto.
  generalize (unlift_lift_eq t); compute; intros Ht; rewrite Ht; clear Ht; auto.
  generalize (unlift_lift_eq t0); compute; intros Ht; rewrite Ht; clear Ht; auto.
  generalize (unlift_lift_eq t1); compute; intros Ht; rewrite Ht; clear Ht; auto.
  generalize (unlift_lift_eq t2); compute; intros Ht; rewrite Ht; clear Ht; auto.
Qed.

Lemma wf_fib_body : (wf_codeval CH (fib_body_type, fib_body_ciseq)).
Proof.
  constructor; intros.
  repeat mcinvClear app_ctype.
  rewrite (unlift_lift_eq t).
  rewrite (unlift_lift_eq t0).
  rewrite (unlift_lift_eq t1).
  rewrite (unlift_lift_eq t2).
  unfold fib_body_ciseq.

  exists 
    (icons (imov r3 r1)
      (icons (imov r1 r2)
        (icons (iadd r2 r2 r3)
          (icons (isubi r0 r0 1)
            (ijd fib_entry_addr
              (tint :: t0 :: t1 :: t2 :: nil))))));
  split; auto.
  repeat constructor.

  repeat constructor; auto.
  eapply wf_ijd; eauto.
  compute; eauto.

  compute; auto.
  unfold ctypeOf, fst, fib_body_type, fib_entry_type.
  repeat (constructor; auto; compute).

  constructor; auto.
  elim r; compute; auto.

  generalize (unlift_lift_eq t0); compute; intros Ht; rewrite Ht; clear Ht; auto.
  generalize (unlift_lift_eq t1); compute; intros Ht; rewrite Ht; clear Ht; auto.
  generalize (unlift_lift_eq t2); compute; intros Ht; rewrite Ht; clear Ht; auto.
Qed.


Lemma wf_ch : wf_codeheap CH.
Proof.
  constructor; intro f.
  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_halt.

  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_fib_entry.

  elim f; clear f; [ intros; inversion H |  intros f _ ].
  apply wf_fib_body.

  intros; inversion H.
Qed.


Definition RF : regfile := fun r => wint 0.
Definition RFT : rftypeF := fun r => tint.

Definition IS : iseq :=
  (icons (imovf r7 halt_addr)
  (icons (imovi r0 10)
  (icons (imovi r1 1)
  (icons (imovi r2 1)
    (ijd fib_entry_addr ( tint::tint::tint::tint::nil)))))).



Goal (wf_progstate (CH, RF, IS)).
Proof.
  apply wf_ps with RFT. 
  apply wf_ch.
  constructor; intro r; elim r; compute; constructor; auto.

  unfold IS.
  eapply wf_imovf; eauto.
  compute; eauto.
  repeat (constructor; auto).
  eapply wf_ijd; eauto.
  compute; eauto.

  unfold ctypeOf, fst; eauto.
  repeat (constructor; compute; auto).
  left; elim r; compute; auto.


Qed.


